<?php
require_once ('LineItem.php');
require_once ('EntryList.php');
require_once ('ExtraItem.php');
require_once ('ProductEntry.php');
require_once (__DIR__. '/../../dbutils.php');
require_once (__DIR__. '/../../products.php');
/*
 * This class is used for reading in a file with the definition
 * of types (drink, meal, etc.) and products.
 * 
 * The file uses the space indenting for assignment to higher level items.
 * If an item has no deeper elements it is considered a product, not
 * a type.
 * 
 * A sample can look like
 * 
 * Type1
 * 		Product A
 * 		Product B
 * 		Type 3
 * 			Product C
 * 			Type 7
 * 				Product D
 * 				Product E
 * 			Type 8
 * 				Product F
 * 		Type 4
 * 			Product G
 * 
 * A type can have additional attributes after "=":
 * 	- nothing: do not run the products in this category over kitchen and supplydesk
 * 			(for example: waiter mixes everything in bar and delivers at same time)
 *  - B: do not run the products in this category, but over supplydesk
 *  		(for example: drinks are mixed by waiter at bar, not in kitchen)
 *  - KB: run all products in this category over kitchen and supplydesk
 *  		(the default, if not overwritten by "type = " in higher category)
 *  - K: over kitchen, not over supplydesk
 *  		(for example: cook delivers product himself)
 */

class TypeAndProductFileManager {
	
	private $entries;   // of type EntryList  (in the beginning all items, later the types)
	private $leafArray;  // later filled by all leafs (= products) when filtered out from $entries;
	private $extras;
	
	private $nextIdOfProdType = 1;
	
	var $dbutils;
	
	function __construct() {
		$this->dbutils = new DbUtils();
	}
	
	/*
	 * Look at the beginning of a line and count the number of spaces or tabs
	 */
	private function intendingOfText($text) {
		$charCounter = 0;
		while (($text[$charCounter] == ' ') || ($text[$charCounter] == "\t")) {
			$charCounter++;
		}
		return $charCounter;
	}
	
	private function startsWith($aText, $needle)
	{
		return $needle === "" || strpos($aText, $needle) === 0;
	}
	
	/*
	 * read in the "Speisekarte.txt"
	 * put the content in the array list "EntryList" as it is
	 * without any modifications
	 */
	private function readFile($fileName) {
		// remove old content if any
		$this->entries = new EntryList();
		
		// first add a dummy so that id and index will equal in the beginning
		// (later ID in DB is also starting from 1)
		$this->entries->add(new LineItem(0,0,"Dummy"));  	// index 0
		
		// now really read the file so that content starts at index 1
		$index = 1;
		$handle = fopen ($fileName, "r");
		while (!feof($handle)) {
			$textline = fgets($handle);
			if(!($this->startsWith($textline,'#'))) {
				$depth = $this->intendingOfText($textline);
				$this->entries->add(new LineItem($depth,$index,$textline));
				$index++;
			}
		}
		fclose ($handle);
	}
	
	private function findNextIdOfProdType($pdo) {
		$index = 1;
		
		$sql = "SELECT id FROM %prodtype% ORDER BY id DESC LIMIT 1";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();
		$row =$stmt->fetchObject();
		if ($row != null) {
			$index = intval($row->id) +1;
		}
		$this->nextIdOfProdType = $index;
	}
	
	public static function removeExtras($pdo) {
		$sql = "DELETE FROM %extrasprods%";
		$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
		$stmt->execute();
		
		$sql = "UPDATE %extras% SET removed=?";
		$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
		$stmt->execute(array('1'));
	}
	
	/*
	 * read in the "Speisekarte.txt"
	* put the content in the array list "EntryList" as it is
	* without any modifications
	*/
	private function parseContent($pdo,$speisekarte) {
		// remove old content if any
		$this->entries = new EntryList();
		$this->extras = array();
	
		self::removeExtras($pdo);

		// get max number of prodtype id -> the old ones will be kept!
		$index = $this->nextIdOfProdType;
		
		$lines = explode("\n", $speisekarte);

		$previousDepth = 0;
		for ($i=0;$i<count($lines);$i++) {
			$textline = $lines[$i];
			$cleanLine = str_replace(" ", "", $textline);
			if (($this->startsWith($textline,'!')) && (strlen($cleanLine) > 0)) {
				$this->extras[] = $textline;
			} else if(!($this->startsWith($textline,'#')) && (strlen($cleanLine) > 0)) {
				$depth = $this->intendingOfText($textline);
				if ($depth > ($previousDepth+1)) {
					return array("status" => "ERROR","code" => PARSE_ERROR,"msg" => PARSE_ERROR_MSG,"line" => $textline);
				} else {
					$previousDepth = $depth;
				}
				$textline = trim($textline);
				$newLineItem = new LineItem($depth,$index,$textline);
				if (is_null($newLineItem)) {
					return array("status" => "ERROR","code" => PARSE_ERROR,"msg" => PARSE_ERROR_MSG,"line" => $textline);
				} else {
					$this->entries->add($newLineItem);
					$index++;
				}
			}
		}
		return array("status" => "OK");
	}
	
	/*
	 * if something is a leaf node it means that it has no
	 * more susequent items with a higher depth.
	 * 
	 * In the sense of the Speisekarte this is a 'product' and
	 * not a 'type'.
	 */
	private function isLeafNode($index) {
		$currentEntry = $this->entries->get($index);
		if ($currentEntry->getType() == ENTRY_TYPE) {
			return false;
		} else {
			return true;
		}
	}
	
	/*
	 * This is like if the entry at index $abIndex would have
	 * been removed from the array. (In fact this will be done
	 * after call of this method probably).
	 * This means that all the entries later will have an index
	 * that has to be decremented by 1. In this case index=id
	 * (the index is stored in the id).
	 * If lower elements have references to the elements with 
	 * decreased indeces, they have to lower their reference as
	 * well.
	 */
	private function reduceIndex($abIndex) {
		// now decrease the index and references if the refer to the reduced part
		$theEntryToRemove = $this->entries->get($abIndex);
		$idOfEntryToRemove = intval($theEntryToRemove->getId());
	    for ($i=$abIndex;$i<$this->entries->size();$i++) {
	    	$entry = $this->entries->get($i);
	    	$theId = intval($entry->getId()); // maybe not necessary going by id
	    	$ref = intval($entry->getReference());
	    	$entry->setId($theId-1);
	    	$theName = $entry->getName();
	    	if ($ref >= $idOfEntryToRemove) {
	    		$entry->setReference($ref-1);
	    	}
	    }
	}
	
	/*
	 * Recursive called function!
	 * 
	 * Find all subitems (with a deeper indenting) and start this
	 * method on these deeper items.
	 * 
	 * This method sets the references, so that it has no return value.
	 */
	private function findAllSubItemsOfIndex($index,$useKitchen,$useSupplyDesk,$kind,$usePrinter) {
		$currentEntry = $this->entries->get($index);
		$currentId = $currentEntry->getId(); // maybe currentId = $index enough?
		$currentDepth = $currentEntry->getDepth();

		// now set the type (leaf -> product, not lead -> type)
		if ($this->isLeafNode($index)) {
			$currentEntry->setType(ENTRY_PRODUCT);
		} else {
			$currentEntry->setType(ENTRY_TYPE);
			$useKitOfEntry = $currentEntry->getUseKitchen();
			$useSupplyOfEntry = $currentEntry->getUseSupplyDesk();
			$kindOfEntry = $currentEntry->getKind();
			$printer = $currentEntry->getPrinter();
			
			if ($useKitOfEntry < 0) {
				$this->entries->get($index)->setUseKitchen($useKitchen);
			} else {
				$useKitchen = $useKitOfEntry;
			}
			if ($useSupplyOfEntry < 0) {
				$this->entries->get($index)->setUseSupplyDesk($useSupplyDesk);
			} else {
				$useSupplyDesk = $useSupplyOfEntry;
			}
			if ($kindOfEntry == KIND_UNDEFINED) {
				$this->entries->get($index)->setKind($kind);
			} else {
				$kind = $kindOfEntry;
			}
			if ($printer < 0) {
				$this->entries->get($index)->setPrinter($printer);
			} else {
				if (!is_null($printer)) {
					$usePrinter = $printer;
				} else {
					$usePrinter = 1;
				}
			}
		}
		
		// now look at all the subsequent items after this item
		for ($nextIndex=($index+1);$nextIndex<$this->entries->size();$nextIndex++) {
			$nextEntry = $this->entries->get($nextIndex);
			$nextId = $nextEntry->getId();
			$nextDepth = $nextEntry->getDepth();
			
			if ($nextDepth == ($currentDepth + 1)) {
				// directly under currentDepth
				$nextEntry->setReference($currentId);
				$this->findAllSubItemsOfIndex($nextIndex,$useKitchen,$useSupplyDesk,$kind,$usePrinter); // recursive!
			} else if ($nextDepth <= $currentDepth) {
				// same depth level or higher - is no sub-element
				break;
			}
		}
	}

	
	/*
	 * Sort out the leafs from the entries. In the entries only the 
	 * types shall be left. The leafs (products) shall go to another
	 * array.
	 */
	private function sortOutLeafs() {
		$this->leafArray = new EntryList(); // is of same type
		for ($index=0;$index < $this->entries->size();$index++) {
			$currentEntry = $this->entries->get($index);
			if ($currentEntry->getType() == ENTRY_PRODUCT) {
				$this->leafArray->add($currentEntry);
				// now modify the references from current index+1 by a ref-1
				$this->reduceIndex($index);
				
				$this->entries->remove($index);
				$index--;
			}
		}
	}
	
	/*
	 * Fill the table prodtype in the DB. The access is done via the 
	 * view name.
	 */
	private function fillProdTypeDbTable ($pdo) {
		for ($i = 0;$i < $this->entries->size(); $i++) {
			$usekitchen = 1; // default
			$usesupplydesk = 1; // default
			$printer = 1; // default
			$theEntry = $this->entries->get($i);
			
			$parts = explode(';', $theEntry->getName(), 2);
			//$parts = explode(';', "123 ; 45", 2);
			$theNameEntry = $parts[0];
			$usekitchen = $theEntry->getUseKitchen();
			$usesupplydesk = $theEntry->getUseSupplyDesk();
			$kind = $theEntry->getKind();
			$printer = $theEntry->getPrinter();
			
			$theProdTypeName = trim($theNameEntry);
			$theRefId = $theEntry->getReference();
			$id = $theEntry->getId();

			$insertSql = "INSERT INTO `%prodtype%` (`id`, `name`, `reference`, `usekitchen`, `usesupplydesk`, `kind`,`printer`) VALUES (?,?,?,?,?,?,?)";
			$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($insertSql));
			try {
				if ($theEntry->getDepth() > 0) {
					$stmt->execute(array($id,$theProdTypeName,$theRefId,$usekitchen,$usesupplydesk,$kind,$printer));
				} else {
					$stmt->execute(array($id,$theProdTypeName,null,$usekitchen,$usesupplydesk,$kind,$printer));
				}
			} catch (Exception $e) {
				return array("status" => "ERROR","code" => PARSE_ERROR,"msg" => PARSE_ERROR_MSG,"line" => $theProdTypeName);
			}
		}
		return array("status" => "OK");
	}
	
	private function fillProductDbTable($pdo,$leafArray) {
		$sortArr = array();
		
		$changedProdIds = array();
		
		for ($i=0;$i < $leafArray->size(); $i++) {
			$theLeafEntry = $leafArray->get($i);
			$product = new ProductEntry();
			$ret = $product->parse($theLeafEntry->getName());
			if ($ret["status"] != "OK") {
				return $ret;
			}
			$priceA = $product->getPriceA();
			$priceB = $product->getPriceB();
			$priceC = $product->getPriceC();
			$unit = $product->getUnit();
			$days = $product->getDays();
			$tax = $product->getTax();
			$taxaustria = trim($product->getTaxAustria());
			if (($taxaustria == "") || ($taxaustria == 0)) {
			    $taxaustria = null;
			}
			$amount = trim($product->getAmount());
			if (trim($amount) == "") {
			    $amount = null;
			}
			$shortName = $product->getShortName();
			$longName = $product->getLongName();
			$prodid = $product->getProdId();
			$available = $product->getAvailable();
			$favorite = 0;
			$prodimageid = $product->getProdImageId();
			$display = $product->getDisplay();
			
			$sql = "SELECT id FROM %prodimages% WHERE id=?";
			$prodimages = CommonUtils::fetchSqlAll($pdo, $sql, array($prodimageid));
			if (count($prodimages) == 0) {
				$prodimageid = null;
			}
			
			$category = $theLeafEntry->getReference();
			
			$sorting = 0;
			if (array_key_exists($category, $sortArr)) {
			    $sorting = $sortArr[$category] + 1;
			    $sortArr[$category] = $sorting;
			} else {
			    $sortArr[$category] = 0;
			}

			$isNewProd = true;
			if (!is_null($prodid)) {
				$sql = "SELECT count(id) as countid FROM %products% WHERE id=?";
				$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
				$stmt->execute(array($prodid));
				$row = $stmt->fetchObject();
				$countid = $row->countid;
				if ($countid > 0) {
					$isNewProd = false;
				}
			}
			if ($isNewProd) {
				$sql = "INSERT INTO `%products%` (`id`, `shortname`, `longname`,  `priceA`,  `priceB`,  `priceC`, `unit`,`days`, `tax`, `taxaustria`, `amount`, `category`,`available`,`favorite`,`sorting`,`prodimageid`,`display`) VALUES (NULL,";
				$sql = $sql . "?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
				$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
				$stmt->execute(array($shortName,$longName,$priceA,$priceB,$priceC,$unit,$days,$tax, $taxaustria, $amount, $category,$available,$favorite,$sorting,$prodimageid,$display));
				$prodid = $pdo->lastInsertId();
				HistFiller::createProdInHist($pdo, $prodid);
			} else {
				
				$sql = "SELECT shortname,longname,priceA,priceB,priceC,unit,days,tax,taxaustria,amount,available,category,prodimageid,display FROM %products% WHERE id=?";
				$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
				$stmt->execute(array($prodid));
				$row = $stmt->fetchObject();
				
				$changed = (($row->shortname != $shortName) 
					|| ($row->longname != $longName) 
					|| ($row->priceA != $priceA)
					|| ($row->priceB != $priceB)
					|| ($row->priceC != $priceC)
					|| ($row->unit != $unit)
					|| ($row->days != $days)
					|| ($row->available != $available)
					|| ($row->tax != $tax)
					|| ($row->taxaustria != $taxaustria)
					|| ($row->prodimageid != $prodimageid)
					|| ($row->display != $display)
					|| ($row->amount != $amount));
				
				if ($changed) {
					$sql = "UPDATE %products% SET shortname=?,longname=?,priceA=?,priceB=?,priceC=?,unit=?,days=?,tax=?,taxaustria=?,amount=?,available=?,sorting=?,prodimageid=?,display=? WHERE id=?";
					$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
					$stmt->execute(array($shortName,$longName,$priceA,$priceB,$priceC,$unit,$days,$tax,$taxaustria,$amount,$available,$sorting,$prodimageid,$display,$prodid));
					$changedProdIds[] = $prodid;
				} else {
				    $sql = "UPDATE %products% SET sorting=? WHERE id=?";
				    $stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
				    $stmt->execute(array($sorting,$prodid));
				}
				
				$sql = "UPDATE %products% SET category=?,removed=? WHERE id=?";
				$stmt = $pdo->prepare(DbUtils::substTableAlias($sql));
				$stmt->execute(array($category,null,$prodid));
			}
		}
		return (array("status" => "OK","changedprods" => $changedProdIds));
	}
	
	/*
	 * Only for test purposes: display entries (=types) and leafs
	 */
	private function outputAll() {
		echo "Types: <br>";
		echo "<table border=1>";
		echo "<tr><th>Index<th>Name<th>Referenz<th>K&uuml;che<th>Bereitstellung</tr>";
		for ($i=1;$i < $this->entries->size(); $i++) {
			$theEntry = $this->entries->get($i);
			if ($theEntry->getReference() > 0) {
				$theRefEntry = $this->entries->get($theEntry->getReference());
				echo "<tr><td>$i<td>" . $theEntry->getName() . "<td>" . $theRefEntry->getName() . "<td>" . $theEntry->getUseKitchen() . "<td>" . $theEntry->getUseSupplyDesk() . "</tr>";
				//echo $i . ":" . $theEntry->getName() . " -> " . $theRefEntry->getName() . "<br>";
			} else {
				echo "<tr><td>$i<td>" . $theEntry->getName() . "<td>null" . "<td>" . $theEntry->getUseKitchen() . "<td>" . $theEntry->getUseSupplyDesk() . "</tr>";
				//echo $i . ":" . $theEntry->getName() . " -> null<br>";
			}
		}
		echo "</table>";
		
		echo "<br>Leafs: <br>";
		for ($i=1;$i < $this->leafArray->size(); $i++) {
			$theEntry = $this->leafArray->get($i);
			if ($theEntry->getReference() > 0) {
				$theRefEntry = $this->entries->get($theEntry->getReference());
				$product = new ProductEntry($theEntry->getName());
				// echo $theEntry->getName() . " -> " . $theRefEntry->getName() . " ---- ";
				echo $product->toString() . " -> " . $theRefEntry->getName() . "<br>";
			} else {
				echo $theEntry->getName() . " -> null<br>";
			}
		}
	}

	
	function manageSpeisekarte($pdo,$content) {
		$this->findNextIdOfProdType($pdo);
		// first remove previous content
		$sql = "UPDATE %products% SET removed='1'";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();

		$sql = "UPDATE `%prodtype%` SET removed='1'";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();
		// now fill in the new content
				
		$ret = $this->parseContent($pdo,$content);
		if ($ret["status"] != "OK") {
			return $ret;
		}

		// now look on the first depth level (0). This level has
		// to be treated individually as staring point because later
		// on it will be iterated recursively
	
		// look at all elements at depth level 0
		
		for ($i=0;$i<$this->entries->size();$i++) {
			$anEntry = $this->entries->get($i);

			if ($anEntry->getDepth() == 0) {
				// highest level
				$this->findAllSubItemsOfIndex($i,1,1,FOOD,null);
			}
		}

		//Debug ausgabe
		for ($i=0;$i<$this->entries->size();$i++) {
			$anEntry = $this->entries->get($i);
			$theId = $anEntry->getId();
			$theName = $anEntry->getName();
			$theRef = $anEntry->getReference();
		}

		$this->sortOutLeafs();
		
		$ret = $this->fillProdTypeDbTable($pdo);
		if ($ret["status"] != "OK") {
			return $ret;
		}
		$ret = $this->fillProductDbTable($pdo,$this->leafArray);
		$changedprodids = $ret["changedprods"]; 

		// now add the extras
		$prodInstance = new Products();

		foreach($this->extras as $anExtraLine) {
			$anExtra = new ExtraItem($pdo,$anExtraLine);
			$prodInstance->createExtraCore($pdo, $anExtra->getName(), $anExtra->getPrice(), $anExtra->getAssignedProdIds());
			
			$changedExtrasProdIds = $anExtra->getAssignedProdIds();
			foreach($changedExtrasProdIds as $anId) {
				if (!in_array($anId, $changedprodids)) {
					$changedprodids[] = $anId;
				}
			}
		}
		
		foreach ($changedprodids as $anId) {
			try {
				HistFiller::updateProdInHist($pdo, $anId);
			} catch (Exception $ex) {
				
			}
		}
		unset($ret["changedprods"]);
		return $ret;
	}
}